#include "y4moutputmodule.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>

#include <string>
#include <vector>
#include <sstream>
#include "libcolorconv.h"

static logT s_log;

struct _MyInstance
{
  _MyInstance()
    : fd(0)
  {
    scaledFb=framebuffer_newInstance();
  }

  ~_MyInstance()
  {
    if (fd) close(fd);
    
    framebuffer_deleteInstance(scaledFb);
  }

  std::string filename;
  int fd;
  size_t x_res;
  size_t y_res;
  std::vector<uint8_t> yuv_buf;
  size_t stream_header_size;
  size_t frame_header_size;
  bool wrote_header;
  FrameBufferType* scaledFb;
};

typedef _MyInstance MyInstance;
typedef _MyInstance* MyInstancePtr;

int init(logT log_function)
{
  s_log = log_function;
  
  return 1;
}

void shutDown(void)
{
}

MyInstance* construct()
{
  return new MyInstance;
}

void destruct(MyInstance* my)
{
  delete my;
}

void update(void* instance)
{
  InstancePtr inst = (InstancePtr) instance;
  MyInstance& my = *inst->my;

  int xsize = trim_int(inst->in_xsize->number, 1, 1400);
  int ysize = trim_int(inst->in_ysize->number, 1, 1080);

  // change size to valid one
  xsize -= xsize%4;
  ysize -= ysize%4;

  // yuv4mpeg2 spec
  assert(xsize%4==0);
  assert(ysize%4==0);

  std::string filename(inst->in_filename->text);
  
  // new stream?
  if (my.fd == 0 || 
      filename != my.filename ||
      xsize != my.x_res || 
      ysize != my.y_res)
    {
      if (my.fd != 0)
	{
	  close(my.fd);
	}
      
      // new stream
      my.x_res=xsize;
      my.y_res=ysize;
      my.filename=filename;
      
      // open new file
      
      // TODO append or new
      my.fd= open(filename.c_str(),O_RDWR|O_CREAT,00666);
      // TODO errorcheck
      
      std::ostringstream stream_header_str;
      stream_header_str << "YUV4MPEG2 ";
      stream_header_str << "W" << my.x_res << " ";
      stream_header_str << "H" << my.y_res << " ";
      stream_header_str << "F30:1 I? A0:0\n";
      
      std::string stream_header(stream_header_str.str());
      my.stream_header_size=stream_header.length();
      
      std::string frame_header("FRAME\n");
      my.frame_header_size=frame_header.length();
      
      my.yuv_buf.resize(my.stream_header_size +
			my.frame_header_size + 
			(my.x_res*my.y_res*6)/4);
      
      // stream_header
      std::copy(stream_header.begin(),
		stream_header.end(),
		my.yuv_buf.begin());
      
      // frame_header
      std::copy(frame_header.begin(),
		frame_header.end(),
		my.yuv_buf.begin() + my.stream_header_size);
      
      // write header next time
      my.wrote_header=false;
    }

  // the framebuffer
  uint32_t* fb;

  if (inst->in_in->xsize != my.x_res || inst->in_in->ysize != my.y_res)
    {
      FrameBufferAttributes  in_attr;
      framebuffer_getAttributes(inst->in_in, &in_attr);
      in_attr.xsize=my.x_res;
      in_attr.ysize=my.y_res;
      framebuffer_convertType(my.scaledFb, inst->in_in , &in_attr);
      
      // use the scaled one
      fb=my.scaledFb->framebuffer;
    }
  else
    {
      // use the original
      fb=inst->in_in->framebuffer;
    }

  // convert
  {
    int pitches[3] ={ my.x_res, my.x_res/2, my.x_res/2 };
    int offsets[3] ={0, my.x_res*my.y_res, 5*(my.x_res*my.y_res)/4};

    cvt_rgb32_to_i420(&my.yuv_buf[my.stream_header_size +my.frame_header_size],
		      fb,
		      my.x_res, my.y_res,
		      pitches, offsets);
    
  }

  // and write  
  {
    uint8_t* start;
    size_t length;
  
    if ( my.wrote_header )
      {
	start= &my.yuv_buf[my.stream_header_size];
	length= my.yuv_buf.size()-my.stream_header_size;
      }
    else
      {
	start= &my.yuv_buf[0];
	length= my.yuv_buf.size();
	my.wrote_header=true;
      }

    // TODO select    
    write(my.fd,start,length);
    // TODO errorcheck
    // TODO partial writes
  }
}

